#include "core/Network.hpp"

#define ENET_IMPLEMENTATION
#include <cassert>
#include <core/HashMap.hpp>
#include <core/Logger.hpp>
#include <core/Utility.hpp>
#include <cstring>

#include <enet.h>

#include "ErrorSimulator.hpp"

using Core::Client;
using Core::InPacket;
using Core::OutPacket;
using Core::Server;

InPacket::InPacket(const void* d, size_t n) :
    data(static_cast<const char*>(d)), size(n), index(0) {
}

bool InPacket::readU8(u8& u) {
    return read(&u, sizeof(u));
}

bool InPacket::readU16(u16& u) {
    if(read(&u, sizeof(u))) {
        return true;
    }
    u = ntohs(u);
    return false;
}

bool InPacket::readU32(u32& u) {
    if(read(&u, sizeof(u))) {
        return true;
    }
    u = ntohl(u);
    return false;
}

bool InPacket::readI8(i8& i) {
    u8 u;
    if(readU8(u)) {
        return true;
    }
    i = static_cast<i8>(static_cast<i32>(u) - static_cast<i32>(128));
    return false;
}

bool InPacket::readI16(i16& i) {
    u16 u;
    if(readU16(u)) {
        return true;
    }
    i = static_cast<i16>(static_cast<i32>(u) - static_cast<i32>(32'768));
    return false;
}

bool InPacket::readI32(i32& i) {
    u32 u;
    if(readU32(u)) {
        return true;
    }
    i = static_cast<i32>(
        static_cast<i64>(u) - static_cast<i64>(2'147'483'648ll));
    return false;
}

bool InPacket::readFloat(float& f) {
    u32 u;
    static_assert(sizeof(u) == sizeof(f), "float and u32 size do not match");
    if(readU32(u)) {
        return true;
    }
    memcpy(&f, &u, sizeof(float));
    return false;
}

size_t InPacket::readString(char* buffer, size_t n) {
    if(n == 0) {
        return 0;
    }
    u16 l;
    if(readU16(l)) {
        return 0;
    }
    size_t end = l;
    char* bufferStart = buffer;
    n--;
    while(n-- > 0 && end > 0) {
        end--;
        u8 u;
        if(readU8(u)) {
            *bufferStart = '\0';
            break;
        }
        *(buffer++) = static_cast<char>(u);
    }
    u8 dummy;
    while(end-- > 0 && !readU8(dummy)) {}
    *buffer = '\0';
    return l;
}

bool InPacket::read(void* buffer, size_t n) {
    if(index + n > size) {
        return true;
    }
    memcpy(buffer, data + index, n);
    index += n;
    return false;
}

void OutPacket::writeU8(u8 u) {
    data.add(u);
}

void OutPacket::writeU16(u16 u) {
    u = htons(u);
    data.add(u);
}

void OutPacket::writeU32(u32 u) {
    u = htonl(u);
    data.add(u);
}

void OutPacket::writeI8(i8 i) {
    writeU8(static_cast<u8>(static_cast<i32>(i) + static_cast<i32>(128)));
}

void OutPacket::writeI16(i16 i) {
    writeU16(static_cast<u16>(static_cast<i32>(i) + static_cast<i32>(32'768)));
}

void OutPacket::writeI32(i32 i) {
    writeU32(static_cast<u32>(
        static_cast<i64>(i) + static_cast<i64>(2'147'483'648)));
}

void OutPacket::writeFloat(float f) {
    u32 u;
    static_assert(sizeof(u) == sizeof(f), "float and u32 size do not match");
    memcpy(&u, &f, sizeof(float));
    writeU32(u);
}

void OutPacket::writeString(const char* s) {
    size_t l = Core::min<size_t>(65'534, strlen(s));
    writeU16(static_cast<u16>(l + 1));
    write(s, l);
    writeU8(0);
}

void OutPacket::write(const void* buffer, size_t n) {
    data.add(buffer, n);
}

static int enetCounter = 0;

static bool addENet() {
    if(enetCounter == 0 && FAIL(enet_initialize() != 0, true)) {
        return true;
    }
    enetCounter++;
    return false;
}

static void removeENet() {
    if(enetCounter > 0 && --enetCounter == 0) {
        enet_deinitialize();
    }
}

static_assert(sizeof(enet_uint16) == sizeof(Core::Port), "port has wrong type");

static void voidVoidDummy() {
}

static void voidInPacketDummy(InPacket&) {
}

struct Client::Data {
    ENetHost* client = nullptr;
    ENetPeer* connection = nullptr;
    Core::OnServerConnect onConnect = voidVoidDummy;
    Core::OnServerDisconnect onDisconnect = voidVoidDummy;
    Core::OnServerPacket onPacket = voidInPacketDummy;
    int connectTicks = 0;
    int connectTimeoutTicks = 0;
    int disconnectTicks = 0;
    int disconnectTimeoutTicks = 0;
};

Client::Client() : data(new Data()) {
}

Client::~Client() {
    stop();
}

bool Client::start() {
    if(data->client != nullptr) {
        REPORT(LogLevel::WARNING, "Client already started");
        return true;
    } else if(addENet()) {
        REPORT(LogLevel::ERROR, "Client cannot initialize enet");
        return true;
    }
    data->client = FAIL(enet_host_create(nullptr, 1, 2, 0, 0), nullptr);
    if(data->client == nullptr) {
        stop();
        REPORT(LogLevel::ERROR, "Cannot create enet client host");
        return true;
    }
    return false;
}

void Client::stop() {
    if(data->connection != nullptr) {
        data->onDisconnect();
        FAIL(
            enet_peer_disconnect_now(data->connection, 0),
            enet_peer_reset(data->connection));
        data->connection = nullptr;
    }
    if(data->client != nullptr) {
        enet_host_destroy(data->client);
        data->client = nullptr;
    }
    removeENet();
    data->connectTicks = 0;
    data->disconnectTicks = 0;
}

bool Client::connect(const char* server, Port port, int timeoutTicks) {
    if(data->client == nullptr) {
        REPORT(LogLevel::WARNING, "Client not started");
        return true;
    } else if(data->connection != nullptr) {
        REPORT(LogLevel::WARNING, "Connection already exists");
        return true;
    }

    ENetAddress address = {};
    enet_address_set_host(&address, server);
    address.port = port;

    data->connection =
        FAIL(enet_host_connect(data->client, &address, 3, 0), nullptr);
    if(data->connection == nullptr) {
        REPORT(LogLevel::ERROR, "Cannot create connection");
        return true;
    }
    data->connectTicks = 1;
    data->connectTimeoutTicks = timeoutTicks;
    return false;
}

void Client::setTimeout(u32 timeout, u32 timeoutMin, u32 timeoutMax) {
    if(data->connection != nullptr) {
        enet_peer_timeout(data->connection, timeout, timeoutMin, timeoutMax);
    }
}

void Client::disconnect(int timeoutTicks) {
    if(data->connection == nullptr) {
        return;
    }
    data->connectTicks = 0;
    enet_peer_disconnect(data->connection, 0);
    data->disconnectTicks = 1;
    data->disconnectTimeoutTicks = timeoutTicks;
}

static ENetPacket* fromBuffer(const Core::Buffer& b, enet_uint8 index) {
    static const enet_uint32 flags[] = {
        ENET_PACKET_FLAG_RELIABLE, 0, ENET_PACKET_FLAG_UNSEQUENCED};
    return enet_packet_create(b.getData(), b.getLength(), flags[index]);
}

void Client::sendPacket(const OutPacket& p, PacketSendMode mode) {
    if(data->client == nullptr || data->connection == nullptr ||
       data->connectTicks >= 0) {
        return;
    }
    enet_uint8 i = static_cast<enet_uint8>(mode);
    enet_peer_send(data->connection, i, fromBuffer(p.data, i));
}

static void tickClientEvents(Client::Data& d) {
    ENetEvent e;
    while(enet_host_service(d.client, &e, 0) >= 0) {
        switch(e.type) {
            case ENET_EVENT_TYPE_CONNECT:
                d.connectTicks = -1;
                d.onConnect();
                break;
            case ENET_EVENT_TYPE_DISCONNECT_TIMEOUT:
            case ENET_EVENT_TYPE_DISCONNECT:
                d.disconnectTicks = 0;
                d.connectTicks = 0;
                d.onDisconnect();
                d.connection = nullptr;
                break;
            case ENET_EVENT_TYPE_RECEIVE: {
                InPacket in(e.packet->data, e.packet->dataLength);
                d.onPacket(in);
                enet_packet_destroy(e.packet);
                break;
            }
            case ENET_EVENT_TYPE_NONE: return;
        }
    }
}

void Client::tick() {
    if(data->client == nullptr) {
        return;
    }
    tickClientEvents(*data);
    if(data->connectTicks >= 1 &&
       ++data->connectTicks > data->connectTimeoutTicks) {
        data->connectTicks = 0;
        disconnect(data->connectTimeoutTicks);
    }
    if(data->disconnectTicks >= 1 &&
       ++data->disconnectTicks > data->disconnectTimeoutTicks) {
        data->disconnectTicks = 0;
        data->onDisconnect();
        if(data->connection != nullptr) {
            enet_peer_reset(data->connection);
            data->connection = nullptr;
        }
    }
}

void Client::setConnectHandler(OnServerConnect oc) {
    data->onConnect = oc == nullptr ? voidVoidDummy : oc;
}

void Client::setDisconnectHandler(OnServerDisconnect od) {
    data->onDisconnect = od == nullptr ? voidVoidDummy : od;
}

void Client::setPacketHandler(OnServerPacket op) {
    data->onPacket = op == nullptr ? voidInPacketDummy : op;
}

void Client::resetHandler() {
    data->onConnect = voidVoidDummy;
    data->onDisconnect = voidVoidDummy;
    data->onPacket = voidInPacketDummy;
}

bool Client::isConnecting() {
    return data->connectTicks >= 1;
}

bool Client::isConnected() {
    return data->connectTicks < 0;
}

static void voidClientDummy(Server&, Core::ClientHandle) {
}

static void voidClientInPacketDummy(Server&, Core::ClientHandle, InPacket&) {
}

struct Server::Data {
    ENetHost* server = nullptr;
    Core::HashMap<Core::ClientHandle, ENetPeer*> clients{};
    Core::ClientHandle idCounter = 1;
    Core::OnClientConnect onConnect = voidClientDummy;
    Core::OnClientDisconnect onDisconnect = voidClientDummy;
    Core::OnClientPacket onPacket = voidClientInPacketDummy;
};

Server::Server() : data(new Data()) {
}

Server::~Server() {
    stop();
}

bool Server::start(Port port, size_t maxClients) {
    if(maxClients <= 0) {
        REPORT(LogLevel::ERROR, "Invalid max client amount");
        return true;
    } else if(data->server != nullptr) {
        REPORT(LogLevel::WARNING, "Server already started");
        return true;
    } else if(addENet()) {
        REPORT(LogLevel::ERROR, "Server cannot initialize enet");
        return true;
    }

    ENetAddress address = {.host = ENET_HOST_ANY, .port = port};
    data->server =
        FAIL(enet_host_create(&address, maxClients, 3, 0, 0), nullptr);
    if(data->server == nullptr) {
        stop();
        REPORT(LogLevel::ERROR, "Cannot create enet server host");
        return true;
    }
    return false;
}

void Server::stop() {
    if(data->server != nullptr) {
        for(auto e : data->clients) {
            enet_peer_reset(e.value);
        }
        enet_host_destroy(data->server);
        data->server = nullptr;
        data->clients.clear();
    }
    removeENet();
}

static void writeId(ENetPeer* peer, Core::ClientHandle id) {
    static_assert(
        sizeof(peer->data) >= sizeof(id), "private data not big enough for id");
    memcpy(&(peer->data), &id, sizeof(id));
}

static Core::ClientHandle getId(ENetPeer* peer) {
    assert(peer->data != nullptr);
    Core::ClientHandle id = -1;
    memcpy(&id, &(peer->data), sizeof(id));
    return id;
}

static void handleConnect(Server& s, Server::Data& d, ENetEvent* e) {
    Core::ClientHandle id = d.idCounter++;
    assert(!d.clients.contains(id));
    d.clients.add(id, e->peer);
    writeId(e->peer, id);
    d.onConnect(s, id);
}

static void handlePacket(Server& s, Server::Data& d, ENetEvent* e) {
    Core::ClientHandle id = getId(e->peer);
    InPacket in(e->packet->data, e->packet->dataLength);
    d.onPacket(s, id, in);
}

static void handleDisconnect(Server& s, Server::Data& d, ENetEvent* e) {
    Core::ClientHandle id = getId(e->peer);
    d.onDisconnect(s, id);
    d.clients.remove(id);
}

void Server::tick() {
    if(data->server == nullptr) {
        return;
    }
    ENetEvent e;
    while(enet_host_service(data->server, &e, 0) >= 0) {
        switch(e.type) {
            case ENET_EVENT_TYPE_CONNECT:
                handleConnect(*this, *data, &e);
                break;
            case ENET_EVENT_TYPE_RECEIVE:
                handlePacket(*this, *data, &e);
                enet_packet_destroy(e.packet);
                break;
            case ENET_EVENT_TYPE_DISCONNECT_TIMEOUT:
            case ENET_EVENT_TYPE_DISCONNECT:
                handleDisconnect(*this, *data, &e);
                break;
            case ENET_EVENT_TYPE_NONE: return;
        }
    }
}

void Server::sendPacketBroadcast(const OutPacket& p, PacketSendMode mode) {
    if(data->server != nullptr) {
        enet_uint8 i = static_cast<enet_uint8>(mode);
        enet_host_broadcast(data->server, i, fromBuffer(p.data, i));
    }
}

void Server::sendPacket(
    ClientHandle clientId, const OutPacket& p, PacketSendMode mode) {
    if(data->server == nullptr) {
        return;
    }
    ENetPeer** peer = data->clients.search(clientId);
    if(peer != nullptr) {
        enet_uint8 i = static_cast<enet_uint8>(mode);
        enet_peer_send(*peer, i, fromBuffer(p.data, i));
    }
}

void Server::setTimeout(
    ClientHandle clientId, u32 timeout, u32 timeoutMin, u32 timeoutMax) {
    if(data->server == nullptr) {
        return;
    }
    ENetPeer** peer = data->clients.search(clientId);
    if(peer != nullptr) {
        enet_peer_timeout(*peer, timeout, timeoutMin, timeoutMax);
    }
}

void Server::disconnectClient(ClientHandle clientId) {
    if(data->server == nullptr) {
        return;
    }
    ENetPeer** peer = data->clients.search(clientId);
    if(peer != nullptr) {
        enet_peer_disconnect(*peer, 0);
    }
}

void Server::setConnectHandler(OnClientConnect oc) {
    data->onConnect = oc == nullptr ? voidClientDummy : oc;
}

void Server::setDisconnectHandler(OnClientDisconnect od) {
    data->onDisconnect = od == nullptr ? voidClientDummy : od;
}

void Server::setPacketHandler(OnClientPacket op) {
    data->onPacket = op == nullptr ? voidClientInPacketDummy : op;
}

void Server::resetHandler() {
    data->onConnect = voidClientDummy;
    data->onDisconnect = voidClientDummy;
    data->onPacket = voidClientInPacketDummy;
}
