#include "core/WindowManager.hpp"

#include <climits>
#include <core/ArrayList.hpp>
#include <core/Clock.hpp>
#include <core/Logger.hpp>
#include <core/Meta.hpp>
#include <core/Unicode.hpp>
#include <core/Utility.hpp>
#include <cstdio>

#include "GLFW.hpp"
#include "core/VulkanWrapper.hpp"

namespace Window = Core::Window;
using Core::Array;

static bool dummyWindowRunHandler(void*) {
    return !Window::shouldClose();
}

static void dummyWindowTickHandler(void*) {
}

static void dummyWindowRenderHandler(void*, float) {
}

static GLFWwindow* window = nullptr;
static Core::IntVector2 size;
static bool sizeChanged = false;
static Window::RunHandler runHandler = dummyWindowRunHandler;
static Window::TickHandler tickHandler = dummyWindowTickHandler;
static Window::RenderHandler renderHandler = dummyWindowRenderHandler;
static void* runHandlerData = nullptr;
static void* tickHandlerData = nullptr;
static void* renderHandlerData = nullptr;
static i64 nanosPerTick = 1;
static Core::Clock fps;
static Core::Clock tps;

static Core::ArrayList<u32, 2048> input = {};
static Core::ArrayList<char, input.getCapacity() * 4 + 1> convertedInput;
static size_t inputCursor = 0;
static size_t inputLimit = input.getCapacity();
static bool inputActive = false;

struct ButtonData {
    char name[31] = {};
    bool released = false;
    int key = 0;
    int downTime = 0;
    int upEvents = 0;
    int downEvents = 0;
    int controllerUp = 0;
    int controllerDown = 0;
};

static Array<ButtonData, 100> buttons = {{{.name = "unknown"}}};
static size_t buttonIndex = 1;
static Array<Window::Button, GLFW_KEY_LAST + 1> keyToButton = {};
static Array<Window::Button, GLFW_GAMEPAD_BUTTON_LAST + 1> gamepadToButton = {};
static Array<Window::Button, GLFW_MOUSE_BUTTON_LAST + 1> mouseToButton = {};
static Core::Vector2 lastMousePosition;
static Core::Vector2 mousePosition;
static int activeController = -1;
static GLFWgamepadstate lastControllerState;

template<size_t N>
static void onButton(Array<Window::Button, N>& map, int key, int action) {
    if(key < 0 || key >= static_cast<int>(N)) {
        return;
    }
    Window::Button b = map[static_cast<size_t>(key)];
    if(b == 0 || b >= buttonIndex) {
        return;
    }
    if(action == GLFW_RELEASE) {
        buttons[b].upEvents++;
    } else if(action == GLFW_PRESS) {
        buttons[b].downEvents++;
    }
}

static void handleInputKey(int key, int action) {
    if(action == GLFW_RELEASE) {
        return;
    }
    switch(key) {
        case GLFW_KEY_BACKSPACE:
            if(input.getLength() >= inputCursor && inputCursor > 0) {
                inputCursor--;
                input.remove(inputCursor);
            }
            break;
        case GLFW_KEY_LEFT: inputCursor -= inputCursor > 0; break;
        case GLFW_KEY_RIGHT:
            inputCursor += inputCursor < input.getLength();
            break;
    }
}

static void onKey(GLFWwindow*, int key, int, int action, int) {
    if(inputActive) {
        handleInputKey(key, action);
    }
    onButton(keyToButton, key, action);
}

static void onChar(GLFWwindow*, u32 codepoint) {
    if(inputActive && input.getLength() < inputLimit) {
        inputCursor += input.putAt(inputCursor, codepoint) != nullptr;
    }
}

static void onResize(GLFWwindow*, int width, int height) {
    sizeChanged = true;
    size[0] = width;
    size[1] = height;
}

static void onMouse(GLFWwindow*, int button, int action, int) {
    onButton(mouseToButton, button, action);
}

static void onMouseMove(GLFWwindow*, double x, double y) {
    mousePosition[0] = static_cast<float>(x);
    mousePosition[1] = static_cast<float>(y);
}

static bool openWindowI(const Window::Options& o) {
    if(!glfwInit()) {
        REPORT(Core::LogLevel::ERROR, "could not initialize GLFW");
        return true;
    }

    glfwDefaultWindowHints();
    glfwWindowHint(GLFW_VISIBLE, false);
    glfwWindowHint(GLFW_RESIZABLE, true);
    glfwWindowHint(GLFW_DECORATED, !o.fullscreen);
    glfwWindowHint(GLFW_DOUBLEBUFFER, true);
    glfwWindowHint(GLFW_CLIENT_API, GLFW_NO_API);

    GLFWmonitor* m = o.fullscreen ? glfwGetPrimaryMonitor() : nullptr;
    window = glfwCreateWindow(o.size[0], o.size[1], o.name, m, nullptr);
    if(window == nullptr) {
        REPORT(Core::LogLevel::ERROR, "could not create window");
        return true;
    }
    size = o.size;
    glfwSetKeyCallback(window, onKey);
    glfwSetCharCallback(window, onChar);
    glfwSetFramebufferSizeCallback(window, onResize);
    glfwSetMouseButtonCallback(window, onMouse);
    glfwSetCursorPosCallback(window, onMouseMove);
    return Core::Vulkan::init();
}

bool Window::open(const Options& o) {
    if(openWindowI(o)) {
        close();
        return true;
    }
    return false;
}

void Window::close() {
    if(window != nullptr) {
        Vulkan::destroy();
        glfwDestroyWindow(window);
        window = nullptr;
    }
    glfwTerminate();
}

void Window::show() {
    glfwShowWindow(window);
}

void Window::trapCursor() {
    glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
}

void Window::freeCursor() {
    glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
}

bool Window::isCursorTrapped() {
    return glfwGetInputMode(window, GLFW_CURSOR) == GLFW_CURSOR_DISABLED;
}

const Core::IntVector2& Window::getSize() {
    return size;
}

bool Window::hasSizeChanged() {
    return sizeChanged;
}

bool Window::shouldClose() {
    return glfwWindowShouldClose(window);
}

void Window::setRunHandler(RunHandler wr, void* data) {
    runHandlerData = data;
    runHandler = wr == nullptr ? dummyWindowRunHandler : wr;
}

void Window::setTickHandler(TickHandler t, void* data) {
    tickHandlerData = data;
    tickHandler = t == nullptr ? dummyWindowTickHandler : t;
}

void Window::setRenderHandler(RenderHandler r, void* data) {
    renderHandlerData = data;
    renderHandler = r == nullptr ? dummyWindowRenderHandler : r;
}

void Window::setNanosPerTick(i64 nanos) {
    nanosPerTick = nanos <= 0 ? 1 : nanos;
}

static bool searchForGamepad() {
    if(activeController != -1) {
        return true;
    }
    for(int i = GLFW_JOYSTICK_1; i <= GLFW_JOYSTICK_LAST; i++) {
        if(glfwJoystickIsGamepad(i)) {
            activeController = i;
            return true;
        }
    }
    return false;
}

static void checkGamepad() {
    GLFWgamepadstate state;
    if(!glfwGetGamepadState(activeController, &state)) {
        activeController = -1;
        return;
    }
    for(size_t i = 0; i < gamepadToButton.getLength(); i++) {
        Window::Button b = gamepadToButton[i];
        if(b == 0 || b >= buttonIndex) {
            return;
        }
        if(!lastControllerState.buttons[i] && state.buttons[i]) {
            buttons[b].controllerDown++;
        } else if(lastControllerState.buttons[i] && !state.buttons[i]) {
            buttons[b].controllerUp++;
        }
    }
    lastControllerState = state;
}

static void endFrame() {
    sizeChanged = false;
    glfwPollEvents();
    if(searchForGamepad()) {
        checkGamepad();
    }
}

static void tickButton(ButtonData* b) {
    bool down = (b->downEvents > 0) || (b->controllerDown > 0);
    bool up = (b->upEvents == b->downEvents) &&
              (b->controllerUp == b->controllerDown);
    if(b->released) {
        b->downTime = 0;
    }
    b->downTime += down;
    b->released = down && up;

    b->downEvents -= b->upEvents;
    b->upEvents = 0;

    b->controllerDown -= b->controllerUp;
    b->controllerUp = 0;
}

static void tick() {
    tps.update();
    ButtonData* b = buttons.begin();
    for(size_t i = 0; i < buttonIndex; i++) {
        tickButton(b++);
    }
    tickHandler(tickHandlerData);
    lastMousePosition = mousePosition;
}

void Window::run() {
    searchForGamepad(); // this is slow the first time
    tps.last = Core::Clock::getNanos();
    fps.last = Core::Clock::getNanos();
    i64 lag = 0;
    while(runHandler(runHandlerData)) {
        lag += fps.update();
        while(lag >= nanosPerTick) {
            lag -= nanosPerTick;
            tick();
        }
        Vulkan::render();
        renderHandler(
            renderHandlerData,
            static_cast<float>(lag) / static_cast<float>(nanosPerTick));
        endFrame();
    }
}

float Window::getTicksPerSecond() {
    return tps.getUpdatesPerSecond();
}

float Window::getFramesPerSecond() {
    return fps.getUpdatesPerSecond();
}

void Window::setInputLimit(size_t limit) {
    inputLimit = Core::min(limit, input.getCapacity() - 1);
    while(input.getLength() > inputLimit) {
        input.removeLast();
    }
    if(inputCursor > inputLimit) {
        inputCursor = inputLimit;
    }
}

void Window::resetInput() {
    input.clear();
    inputCursor = 0;
}

void Window::enableInput() {
    inputActive = true;
}

void Window::disableInput() {
    inputActive = false;
}

bool Window::isInputEnabled() {
    return inputActive;
}

void Window::fillInput(const char* s) {
    resetInput();
    input.clear();
    while(*s != 0) {
        UTF8 u;
        u.length = getUTF8Length(static_cast<u8>(*s));
        u.data[0] = static_cast<u8>(*(s++));
        for(size_t i = 1; i < u.length && *s != 0; i++) {
            u.data[i] = static_cast<u8>(*(s++));
        }
        input.add(convertUTF8toUnicode(u));
    }
    while(input.getLength() > inputLimit) {
        input.removeLast();
    }
}

size_t Window::getInputCursor() {
    return inputCursor;
}

void Window::setInputCursor(size_t index) {
    if(index > input.getLength()) {
        inputCursor = input.getLength();
    } else {
        inputCursor = index;
    }
}

const char* Window::getInputString() {
    convertedInput.clear();
    for(u32 c : input) {
        UTF8 u = convertUnicodeToUTF8(c);
        for(size_t i = 0; i < u.length; i++) {
            convertedInput.put(u.data[i]);
        }
    }
    if(convertedInput.put(0) == nullptr) {
        convertedInput.removeLast();
        convertedInput.put(0);
    }
    return convertedInput.begin();
}

Window::Button Window::addButton(const char* name) {
    if(buttonIndex >= buttons.getLength()) {
        return 0;
    }
    Button b = buttonIndex++;
    snprintf(buttons[b].name, sizeof(buttons[b].name), "%s", name);
    return b;
}

void Window::bindKeyToButton(Button b, Button key) {
    if(key < keyToButton.getLength()) {
        keyToButton[key] = b;
    }
}

void Window::bindGamepadToButton(Button b, Button gamepadButton) {
    if(gamepadButton < gamepadToButton.getLength()) {
        gamepadToButton[gamepadButton] = b;
    }
}

void Window::bindMouseToButton(Button b, Button mouseButton) {
    if(mouseButton < mouseToButton.getLength()) {
        mouseToButton[mouseButton] = b;
    }
}

Core::Vector2 Window::getLastMousePosition() {
    return lastMousePosition;
}

Core::Vector2 Window::getMousePosition() {
    return mousePosition;
}

Core::Vector2 Window::getLeftGamepadAxis() {
    return {
        lastControllerState.axes[GLFW_GAMEPAD_AXIS_LEFT_X],
        lastControllerState.axes[GLFW_GAMEPAD_AXIS_LEFT_Y]};
}

Core::Vector2 Window::getRightGamepadAxis() {
    return {
        lastControllerState.axes[GLFW_GAMEPAD_AXIS_RIGHT_X],
        lastControllerState.axes[GLFW_GAMEPAD_AXIS_RIGHT_Y]};
}

float Window::getLeftGamepadTrigger() {
    return lastControllerState.axes[GLFW_GAMEPAD_AXIS_LEFT_TRIGGER];
}

float Window::getRightGamepadTrigger() {
    return lastControllerState.axes[GLFW_GAMEPAD_AXIS_RIGHT_TRIGGER];
}

static const ButtonData& getButton(Window::Button id) {
    return buttons[id >= buttonIndex ? 0 : id];
}

bool Window::isButtonDown(Button b) {
    return getButton(b).downTime > 0;
}

int Window::getButtonDownTime(Button b) {
    return getButton(b).downTime;
}

bool Window::wasButtonReleased(Button b) {
    return getButton(b).released;
}

const char* Window::getButtonName(Button b) {
    return getButton(b).name;
}

GLFWwindow* Window::get() {
    return window;
}
